#!/usr/bin/env python
#
# savitch.py
#
# Implementation of the DSPACE((log n)^2) algorithm for STCON
#
# Syntax:
#
#   python savitch.py n_nodes link_probability seed method
#
# with
#   n_nodes: number of nodes in the graph
#   link_probability: probability of generating a link between any pair of nodes
#   seed: random number generator seed
#   method: linear or savitch
#
# The programs generates a random connectivity matrix, then
# tries to find a path from the first to the last node with two methods:
#  - "linear": search for a path node by node
#  - "savitch": recursively split the search in half by searching a suitable midpoint.
#
# The program uses the number of recursive calls as a measure of time, and the maximum
# recursion depth as a measure of space.
#
# For example, compare:
#   python savitch.py 500 .01 1 linear
#   python savitch.py 500 .01 1 savitch
# and note how the 'savitch' strategy uses a very low recursion depth (logarithmic in n_nodes)
# but a much higher number of recursive calls. Conversely, the 'linear' strategy uses a linear
# recursion depth (equal to n_nodes), but a much lower number of calls.
#
# Observe that the only difference between the two strategies is the division into subproblems in
# the recursive call.
#
# Note that the 'linear' strategy is still suboptimal: a lot could be improved, e.g. avoid
# revisiting nodes.

##############################
#
# Modules

import sys # used by input and output
import random # self-explanatory...

##############################
#
# Input parameters

n_nodes = int(sys.argv[1])
link_probability = float(sys.argv[2])
seed  = int(sys.argv[3])
method = sys.argv[4]
use_savitch = method == 'savitch'

sys.stderr.write('Input parameters\n')
sys.stderr.write('Nodes:\t\t%10d\n' % n_nodes)
sys.stderr.write('Link prob.:\t%10f\n' % link_probability)
sys.stderr.write('Seed:\t\t%10d\n' % seed)
sys.stderr.write('Method:\t\t%10s\n' % method)

####################################
#
# Connectivity matrix

sys.stderr.write('\nGenerating random connectivity matrix...\n')
random.seed(seed)
M = [
    [
        i != j and random.random() < link_probability
        for j in range(n_nodes)
    ]
    for i in range(n_nodes)
]

####################################
#
# The algorithm

# A few more globals:
# keep a record of the maximum recursion depth and of the total number of recursive calls
maxdepth = 0
calls = 0

# Path-searching function.
# Parameters:
#   v: source node
#   w: target node
#   l: maximum path length
#   depth: auxiliary parameter used to keep track of the recursion depth
def path_exists(v, w, l = n_nodes, depth = 1):
    # Use the relevant global variables to avoid passing them as parameters
    global M, use_savitch, maxdepth, calls

    # Update the number of recursive calls
    calls += 1
    # Update the maximum recursion depth if needed
    if depth > maxdepth:
        maxdepth = depth

    # Uncomment the following line to keep track of the execution (it slows down the code):
    # sys.stderr.write('\rcalls: %12d\tdepth: %5d\tmaxdepth: %5d' % (calls, depth, maxdepth))

    # Recursion base: l=0 means that length has been exhausted
    if l == 0:
        return False
    # Recursion base: if l=1, need direct connection
    if l == 1:
        return M[v][w]

    # The only difference between linear and savitch execution is how to split into subproblems:
    # if linear, split is l -> (1, l-1)
    # if savitch, split is l-> (l/2, l/2)
    l1 = int(l/2) if use_savitch else 1

    # Iterate through candidate nodes for the split
    for v1 in range(len(M)):
        # Recursively check if the split works
        if path_exists(v, v1, l1, depth+1) and path_exists(v1, w, l-l1, depth+1):
            return True
    # If no suitable split node could be found, no path exists between v and w
    return False

#############################################
#
# Search a path from the first (0) and the last (n_nodes-1) node of the graph

sys.stderr.write('Searching path...\n')
result = path_exists(0, n_nodes-1)

# Report the results

sys.stderr.write('\nResults\n')
sys.stdout.write('Calls:\t\t%10d\n' % calls)
sys.stdout.write('Max depth:\t%10d\n' % maxdepth)
sys.stdout.write('Path exists:\t%10s\n' % result)
